import datetime
from typing import overload

import flanaapis.geolocation.functions
from flanaapis.scraping import google_weather_scraper
from flanaapis.weather.models import DayWeather, InstantWeather


def create_instant_weather_by_data(data: dict, timezone: datetime.timezone, days_offset=0) -> InstantWeather:
    now = datetime.datetime.now(timezone) + datetime.timedelta(days=days_offset)
    hour = datetime.datetime.strptime(data['datetime'].split()[-1], '%H:%M').hour

    return InstantWeather(
        date_time=datetime.datetime(now.year, now.month, now.day, hour, tzinfo=timezone),
        description=data.get('weather'),
        humidity=data.get('humidity'),
        precipitation_probability=data.get('precip_prob'),  # %
        temperature=data.get('temp'),
        temperature_feel=data.get('temp'),
        wind_speed=data.get('wind_speed') or data.get('wind')  # km/h
    )


@overload
async def get_day_weathers_by_place(place_query: str, timezone: datetime.timezone = None) -> tuple[InstantWeather, list[DayWeather]]:
    pass


@overload
async def get_day_weathers_by_place(latitude: float, longitude: float, timezone: datetime.timezone = None) -> tuple[InstantWeather, list[DayWeather]]:
    pass


async def get_day_weathers_by_place(latitude: float | str, longitude: float = None, timezone: datetime.timezone = None) -> tuple[InstantWeather | None, list[DayWeather] | None]:
    latitude, longitude = await flanaapis.geolocation.functions.ensure_coordinates(latitude, longitude)
    place = await flanaapis.geolocation.open_street_map.find_place(f'{latitude}, {longitude}')

    # noinspection PyBroadException
    try:
        api_weather_data = google_weather_scraper.get_forecast(str(place))
    except BaseException:
        return None, None

    if not timezone:
        if api_timezone_data := await flanaapis.geolocation.functions.find_timezone(latitude, longitude):
            timezone = datetime.timezone(datetime.timedelta(seconds=api_timezone_data['gmtOffset']))
        else:
            timezone = datetime.timezone(datetime.timedelta())

    current_weather = create_instant_weather_by_data(api_weather_data['weather_now'], timezone)
    current_weather.date_time = datetime.datetime.now(timezone)

    day_weathers = []

    last_day_name = ''
    days_offset = -1
    day_weather = None
    for hour_data in api_weather_data['hourly_forecast']:
        if last_day_name != (day_name := hour_data['datetime'].split()[0]):
            last_day_name = day_name
            days_offset += 1
            try:
                day_data = api_weather_data['next_days'][days_offset]
            except IndexError:
                day_data = {}
            day_weathers.append(
                day_weather := DayWeather(
                    date=datetime.datetime.now(timezone).date() + datetime.timedelta(days=days_offset),
                    timezone=timezone,
                    min_temperature=day_data.get('min_temp'),
                    max_temperature=day_data.get('max_temp')
                )
            )

        day_weather.instant_weathers.append(create_instant_weather_by_data(hour_data, timezone, days_offset))

    return current_weather, day_weathers



__all__ = ['clear_past_precipitation_probability', 'get_day_weathers_by_place']

import datetime
from typing import overload

import flanautils

from flanaapis.geolocation import functions
from flanaapis.geolocation.models import Place
from flanaapis.weather import google, open_weather_map, visual_crossing
from flanaapis.weather.models import DayWeather, InstantWeather


def clear_past_precipitation_probability(day_weathers: list[DayWeather], timezone: datetime.timezone):
    for day_weather in day_weathers:
        for instant_weather in day_weather.instant_weathers:
            if instant_weather.date_time < datetime.datetime.now(timezone):
                instant_weather.precipitation_probability = None


@overload
async def get_day_weathers_by_place(place: Place, ratios: list[float] = None) -> tuple[InstantWeather, list[DayWeather]]:
    pass


@overload
async def get_day_weathers_by_place(place_query: str, ratios: list[float] = None) -> tuple[InstantWeather, list[DayWeather]]:
    pass


@overload
async def get_day_weathers_by_place(latitude: float, longitude: float, ratios: list[float] = None) -> tuple[InstantWeather, list[DayWeather]]:
    pass


async def get_day_weathers_by_place(latitude: float | str, longitude: float = None, ratios: list[float] = None) -> tuple[InstantWeather, list[DayWeather]]:
    def get_timezone(current_weather, days_weathers) -> datetime.timezone | None:
        try:
            return current_weather.date_time.tzinfo
        except AttributeError:
            try:
                return days_weathers[0].timezone
            except (IndexError, TypeError):
                return None

    latitude, longitude = await functions.ensure_coordinates(latitude, longitude)

    open_current_weather, open_day_weathers = await open_weather_map.get_day_weathers_by_place(latitude, longitude)
    vc_current_weather, vc_day_weathers = await visual_crossing.get_day_weathers_by_place(latitude, longitude)
    timezone = get_timezone(open_current_weather, open_day_weathers) or get_timezone(vc_current_weather, vc_day_weathers)
    google_current_weather, google_day_weathers = await google.get_day_weathers_by_place(latitude, longitude, timezone)

    all_day_weathers = [day_weathers for day_weathers in (open_day_weathers, vc_day_weathers, google_day_weathers) if day_weathers]

    final_day_weathers = []
    if len(all_day_weathers) == 1:
        final_day_weathers = all_day_weathers[0]
    else:
        first_date = sorted(day_weathers[0].date for day_weathers in all_day_weathers)[0]
        last_date = sorted(day_weathers[-1].date for day_weathers in all_day_weathers)[-1]
        date = first_date
        while date <= last_date:
            all_day_weather = [day_weather for day_weathers in all_day_weathers if (day_weather := flanautils.find(day_weathers, condition=lambda day_weather: day_weather.date == date))]
            if len(all_day_weather) == 1:
                final_day_weathers.append(all_day_weather[0])
            else:
                final_day_weathers.append(DayWeather.mean(all_day_weather, ratios))
            date = date + datetime.timedelta(days=1)

    return InstantWeather.mean((open_current_weather, vc_current_weather, google_current_weather), ratios), final_day_weathers



import json
import re

import requests
from bs4 import BeautifulSoup as bs

"""
Part of this code is based on Dniamir's work https://github.com/dniamir/GoogleWeather.
And another part was made possible because of Andrej Kesely answer on https://stackoverflow.com/.
"""

URL = "https://www.google.com/search?hl=es&lr=lang_es&ie=UTF-8&q=weather"
HEADER = {
    "User-Agent": "Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:89.0) Gecko/20100101 Firefox/89.0",
    "Language": "es-ES,es;q=0.5",
}


def mph_to_kmph(mph):
    # Converts Miles per Hour to Kilometers per Hour
    return round(mph * 1.6, 1)


def kmph_to_mph(kmph):
    # Converts Kilometers per Hour to Miles per Hour
    return round(kmph / 1.6, 1)


def f_to_c(f):
    # Converts Farenheit to Celsius
    return round((f - 32) * (5 / 9), 1)


def c_to_f(c):
    # Converts Celsius to Farenheit
    return round(c * (9 / 5) + 32, 1)


def _get_soup(header, url):
    """This functions simply gets the header and url, creates a session and
       generates the "soup" to pass to the other functions.

    Args:
        header (dict): The header parameters to be used in the session.
        url (string): The url address to create the session.

    Returns:
        bs4.BeautifulSoup: The BeautifoulSoup object.
    """

    # Try to read data from URL, if it fails, return None
    try:
        session = requests.Session()
        session.headers["User-Agent"] = header["User-Agent"]
        session.headers["Accept-Language"] = header["Language"]
        session.headers["Content-Language"] = header["Language"]
        html = session.get(url)
        return bs(html.text, "html.parser")
    except:
        print(f"ERROR: Unable to retrieve data from {url}")
        return None


def _get_weather_now(soup, output_units):
    """Gets the current weather conditions.

    Args:
        soup (bs4.BeautifulSoup): The BeautifoulSoup object.
        output_units (dict): A dictionary contatining "temp" key, which can be
                             "c" for Celsius or "f" for Farenheit and a "speed"
                             key, which can be "km/h" for Kilometers Per Hour
                             or "mph" for Miles Per Hour.

    Returns:
        dict: A dictionary containing the current "Temperature", "Datetime",
                "Weather Condition", "Precipitation Probability", "Humidity"
                and "Wind Speed".
    """

    # Create a dictionary to store the output data
    data = dict()

    # Get the region output from Google
    region = soup.find("div", attrs={"id": "wob_loc"}).text

    # Get Weather Data
    data["temp"] = float(soup.find("span", attrs={"id": "wob_tm"}).text)
    data["datetime"] = soup.find("div", attrs={"id": "wob_dts"}).text
    data["weather"] = soup.find("span", attrs={"id": "wob_dc"}).text
    data["precip_prob"] = float(
        soup.find("span", attrs={"id": "wob_pp"}).text.replace("%", "")
    )
    data["humidity"] = float(
        soup.find("span", attrs={"id": "wob_hm"}).text.replace("%", "")
    )
    data["wind"] = soup.find("span", attrs={"id": "wob_ws"}).text

    # Autodetect and convert "Temperature" and "Wind Speed" units
    if "km/h" in data["wind"]:
        data["wind"] = float(data["wind"].replace("km/h", ""))

        if output_units["speed"] == "mph":
            data["wind"] = kmph_to_mph(data["wind"])

        if output_units["temp"] == "f":
            data["temp"] = c_to_f(data["temp"])

        input_units = "metric"
    else:
        data["wind"] = float(data["wind"].replace("mph", ""))

        if output_units["speed"] == "kph":
            data["wind"] = mph_to_kmph(data["wind"])

        if output_units["temp"] == "c":
            data["temp"] = f_to_c(data["temp"])

        input_units = "imperial"

    return input_units, region, data


def _get_next_days(soup, input_units, output_units):
    """Gets a summary of the next 8 days (including today).

    Args:
        soup (bs4.BeautifulSoup): The BeautifoulSoup object.
        input_units (string): The autodetected input units, they can be either
                              "metric" or "imperial".
        output_units (dict): A dictionary contatining "temp" key, which can be
                             "c" for Celsius or "f" for Farenheit and a "speed"
                             key, which can be "km/h" for Kilometers Per Hour
                             or "mph" for Miles Per Hour.

    Returns:
        list: A list of the next 8 days (including today) containing "Weather
              Condition", "Day Name", "Minimum Temperature" and "Maximum
              Temperature" for each day.
    """
    # Create a list to store the output data
    data = list()

    # Extract data from soup
    days = soup.find("div", attrs={"id": "wob_dp"})

    # Iterate over every single day
    for day in days.findAll("div", attrs={"class": "wob_df"}):
        day_name = day.find("div").attrs["aria-label"]
        weather = day.find("img").attrs["alt"]
        temp = day.findAll("span", {"class": "wob_t"})

        # Get the right data for the chosen output units
        if input_units == "metric":
            if output_units["temp"] == "c":
                max_temp = float(temp[0].text)
                min_temp = float(temp[2].text)
            else:
                max_temp = float(temp[1].text)
                min_temp = float(temp[3].text)
        else:
            if output_units["temp"] == "c":
                max_temp = float(temp[1].text)
                min_temp = float(temp[3].text)
            else:
                max_temp = float(temp[0].text)
                min_temp = float(temp[2].text)

        # Append the values to the output list
        data.append(
            {
                "day": day_name,
                "weather": weather,
                "max_temp": max_temp,
                "min_temp": min_temp,
            }
        )

    return data


def _get_wind(soup, output_units):
    """Wind Direction and Wind Bearing must be retrieved in a "special" manner,
       Google provides a 15 day forecast with 3 hour intervals containing Wind
       Speed, Wind Direction, Datetime and Wind Bearing. This function extracts
       this data.

    Args:
        soup (bs4.BeautifulSoup): The BeautifoulSoup object
        output_units (dict): A dictionary contatining "temp" key, which can be
                             "c" for Celsius or "f" for Farenheit and a "speed"
                             key, which can be "km/h" for Kilometers Per Hour
                             or "mph" for Miles Per Hour.

    Returns:
        list: A list containing one entry for every 3 hour period, each entry
              is another list, composed of wind speed, wind direction,
              datetime and wind bearing.
    """
    # Extracting the data from the "soup"
    wind = str(soup.find("div", attrs={"id": "wob_wg", "class": "wob_noe"}))
    data = re.findall(
        r'"(\d+ [\w\/]+) \w+ (\w+) (\w+-*\w*,* [0-9:]+\s*\w*)" class="wob_t" style="display:inline;text-align:right">\d+ [\w\/]+<\/span><span aria-label="\d+ [\w\/]+ \w+-*\w*,* [0-9:]+\s*\w*" class="wob_t" style="display:none;text-align:right">\d+ [\w\/]+<\/span><\/div><div style="[\w-]+:\d+"><\/div><img alt="\d+ [\w\/]+ \w+ \w+" aria-hidden="true" src="\/\/ssl.gstatic.com\/m\/images\/weather\/\w+.\w+" style="transform-origin:\d+% \d+%;transform:rotate\((\d+)\w+\)',
        wind,
    )

    # Extracting the input values units
    if "km/h" in data[0][0]:
        input_units = "metric"
    else:
        input_units = "imperial"

    # Iterating over data to post process it
    for idx, _ in enumerate(data):
        data[idx] = list(data[idx])

        # Removing the offset from Wind Bearing
        data[idx][3] = int(data[idx][3]) - 90

        # Extracting numerical values and converting units, if necessary
        if input_units == "metric":
            data[idx][0] = float(data[idx][0].replace("km/h", ""))

            if output_units["speed"] == "mph":
                data[idx][0] = kmph_to_mph(data[idx][0])

        else:
            data[idx][0] = float(data[idx][0].replace("mph", ""))

            if output_units["speed"] == "km/h":
                data[idx][0] = mph_to_kmph(data[idx][0])

    return data


def _get_hourly_forecast(header, url, output_units):
    """This functions extracts hourly forecast data for the next 15 days.

    Args:
        header (dict): The header parameters to be used in the session.
        url (string): The url address to create the session.
        output_units (dict): A dictionary contatining "temp" key, which can be
                            "c" for Celsius or "f" for Farenheit and a "speed"
                            key, which can be "km/h" for Kilometers Per Hour
                            or "mph" for Miles Per Hour.

    Returns:
        list: A list containing one dictionary for every 1 hour period. Each
              dictionary has "Datetime", "Humidity", "Precipitation Probability"
              "Temperature", "Weather Condition" and "Wind Speed".
    """
    # Build the header
    header = {"User-Agent": header["User-Agent"]}

    # Request and extract the data from the url
    text = requests.get(url, headers=header).text
    data_in = re.search(r"pmc='({.*?})'", text).group(1)
    data_in = json.loads(data_in.replace(r"\x22", '"').replace(r'\\"', r"\""))

    # Create a list to store the output data
    data_out = list()

    # Iterate over each entry
    for entry_in in data_in["wobnm"]["wobhl"]:

        # Create a dictionary to store the output entry
        entry_out = dict()

        # Pull data from the input entry into the output entry
        entry_out["datetime"] = entry_in["dts"]
        entry_out["weather"] = entry_in["c"]
        entry_out["humidity"] = float(entry_in["h"].replace("%", ""))
        entry_out["precip_prob"] = float(entry_in["p"].replace("%", ""))

        # Get the right data for the chosen output units
        if output_units["temp"] == "c":
            entry_out["temp"] = float(entry_in["tm"])
        else:
            entry_out["temp"] = float(entry_in["ttm"])

        if output_units["speed"] == "km/h":
            entry_out["wind_speed"] = float(entry_in["ws"].replace("km/h", ""))
        else:
            entry_out["wind_speed"] = float(entry_in["tws"].replace("mph", ""))

        # Append the output entry to the output list
        data_out.append(entry_out)

    return data_out


def get_forecast(region, output_units={"temp": "c", "speed": "km/h"}):
    """This is the wrapper that calls the other functions and joins the data into
       one output.

    Args:
        header (dict): The header parameters to be used in the session.
        url (string): The url address to create the session.
        region (string): The desired region to get the weather forecast.
        output_units (dict): A dictionary contatining "temp" key, which can be
                            "c" for Celsius or "f" for Farenheit and a "speed"
                            key, which can be "km/h" for Kilometers Per Hour
                            or "mph" for Miles Per Hour.

    Returns:
        dict: A dictionary containing all the data extracted by the other
              functions
    """
    # Build url and get the "soup" from it
    url = f"{URL}+{region.replace(' ', '+')}"
    soup = _get_soup(HEADER, url)

    # Create a dictionary to store the output data
    data = dict()

    # Check if we got a soup to work with
    if soup:
        input_units, data["region"], data["weather_now"] = _get_weather_now(
            soup, output_units
        )
        data["next_days"] = _get_next_days(soup, input_units, output_units)
        data["wind"] = _get_wind(soup, output_units)

    data["hourly_forecast"] = _get_hourly_forecast(HEADER, url, output_units)

    return data



from fastapi import APIRouter

from flanaapis.weather import functions

router = APIRouter()


@router.get("/weather")
async def weather(latitude: float, longitude: float):
    current_weather, day_weathers = await functions.get_day_weathers_by_place(latitude, longitude)

    for i, day_weather in enumerate(day_weathers):
        for j, instant_weather in enumerate(day_weather.instant_weathers):
            day_weather.instant_weathers[j] = instant_weather.to_dict()
        day_weathers[i] = day_weather.to_dict()

    return {
        'current_weather': current_weather.to_dict(),
        'day_weathers': day_weathers
    }



from __future__ import annotations  # todo0 remove in 3.11

import datetime
from dataclasses import dataclass, field
from enum import auto
from typing import Any, Iterable, Sequence

from flanautils import FlanaBase, FlanaEnum, MeanBase


class DayPhases(FlanaEnum):
    NIGHT = 2
    MORN = 9
    DAY = 14
    EVE = 19


class PrecipitationType(FlanaEnum):
    FREEZINGRAIN = auto()
    ICE = auto()
    RAIN = auto()
    SNOW = auto()


@dataclass(unsafe_hash=True)
class Precipitation(FlanaBase):
    type_: PrecipitationType
    start_date: datetime.datetime
    end_date: datetime.datetime
    volume: float


@dataclass(unsafe_hash=True)
class InstantWeather(MeanBase, FlanaBase):
    date_time: datetime.datetime = None
    clouds: float = field(default=None, metadata={'unit': '%'})
    description: str = None
    dew_point: float = field(default=None, metadata={'unit': 'ºC'})
    humidity: float = field(default=None, metadata={'unit': '%'})
    icon: Any = None
    precipitation_probability: float = field(default=None, metadata={'unit': '%'})
    pressure: float = field(default=None, metadata={'unit': 'hPa'})
    rain_volume: float = field(default=None, metadata={'unit': 'mm'})
    snow_volume: float = field(default=None, metadata={'unit': 'mm'})
    temperature: float = field(default=None, metadata={'unit': 'ºC'})
    temperature_feel: float = field(default=None, metadata={'unit': 'ºC'})
    uvi: float = None
    visibility: float = field(default=None, metadata={'unit': 'km'})
    wind_degrees: float = None
    wind_gust: float = field(default=None, metadata={'unit': 'km/h'})
    wind_speed: float = field(default=None, metadata={'unit': 'km/h'})

    def _dict_repr(self) -> Any:
        self_vars = super()._dict_repr()
        for k, v in self_vars.items():
            if isinstance(v, datetime.datetime):
                self_vars[k] = v.timestamp()
            elif isinstance(v, datetime.timezone):
                self_vars[k] = str(v)

        return self_vars

    @classmethod
    def mean(
        cls,
        objects: Sequence,
        ratios: list[float] = None,
        attribute_names: Iterable[str] = ('clouds', 'dew_point', 'humidity', 'precipitation_probability', 'pressure', 'rain_volume', 'snow_volume', 'temperature', 'temperature_feel', 'uvi', 'visibility', 'wind_degrees', 'wind_gust', 'wind_speed')
    ) -> InstantWeather:
        if not objects:
            return InstantWeather()

        # noinspection PyTypeChecker
        instant_weather: InstantWeather = super().mean(objects, ratios, attribute_names)
        for object_ in objects:
            if object_:
                instant_weather.date_time = object_.date_time

        return instant_weather

    def merge(self, other: InstantWeather, left_priority=True) -> InstantWeather:
        for k, v in vars(self).items():
            if v is None or not left_priority:
                setattr(self, k, getattr(other, k, None))
        return self


@dataclass(unsafe_hash=True)
class DayWeather(MeanBase, FlanaBase):
    date: datetime.date = None
    timezone: datetime.timezone = None
    sunrise: datetime.datetime = None
    sunset: datetime.datetime = None
    min_temperature: float = None
    max_temperature: float = None
    instant_weathers: list[InstantWeather] = field(default_factory=list)

    def _dict_repr(self) -> Any:
        self_vars = super()._dict_repr()
        for k, v in self_vars.items():
            if isinstance(v, datetime.date):
                self_vars[k] = datetime.datetime(v.year, v.month, v.day).timestamp()
            elif isinstance(v, datetime.timezone):
                self_vars[k] = str(v)

        return self_vars

    def distribute_precipitation_volume(self, precipitations: Iterable[Precipitation]):
        def distribute_precipitation_volume_(precipitation_type: PrecipitationType, precipitations_: Iterable[Precipitation]):
            grouped_precipitations: dict[int, list[Precipitation]] = {}
            for hours in range(1, 25):
                grouped_precipitations[hours] = [precipitation.copy() for precipitation in precipitations_ if (precipitation.end_date - precipitation.start_date).total_seconds() / 3600 == hours]

            hourly_precipitations: list[Precipitation] = grouped_precipitations[1]
            for hours in range(2, 25):
                for precipitation in grouped_precipitations[hours]:
                    start_hour = precipitation.start_date.hour
                    start_hours = list(range(start_hour, min(start_hour + hours, 24)))

                    for hourly_precipitation in hourly_precipitations:
                        if hourly_precipitation.start_date.hour in start_hours:
                            start_hours.remove(hourly_precipitation.start_date.hour)
                            precipitation.volume -= hourly_precipitation.volume
                    if start_hours:
                        hourly_volume = precipitation.volume / len(start_hours)
                    else:
                        continue
                    for start_hour in start_hours:
                        start_date = precipitation.start_date.replace(hour=start_hour)
                        hourly_precipitations.append(Precipitation(precipitation_type, start_date, start_date + datetime.timedelta(hours=1), hourly_volume))
            hourly_precipitations.sort(key=lambda precipitation: precipitation.start_date)

            attribute_name = f'{precipitation_type.name.lower()}_volume'
            for precipitation in hourly_precipitations:
                if not precipitation.volume:
                    return

                if not (instant_weather := next(iter(self.get_instant_weathers_by_hour(precipitation.start_date.hour)), None)):
                    instant_weather = InstantWeather(precipitation.start_date)
                    self.instant_weathers.append(instant_weather)
                setattr(instant_weather, attribute_name, precipitation.volume)

            self.instant_weathers.sort(key=lambda instant_weather_: instant_weather_.date_time)

        precipitations = [precipitation for precipitation in precipitations if precipitation.start_date.date() == self.date]
        temp_precipitations = []
        for precipitation in precipitations:
            next_day_first_hour_date_time = (precipitation.start_date + datetime.timedelta(days=1)).replace(hour=0)
            if next_day_first_hour_date_time < precipitation.end_date:
                precipitation_hours = (precipitation.end_date - precipitation.start_date).total_seconds() / 3600
                hours_until_next_day = (next_day_first_hour_date_time - precipitation.start_date).total_seconds() / 3600
                next_day_precipitation_hours = precipitation_hours - hours_until_next_day
                temp_precipitations.append(Precipitation(precipitation.type_, precipitation.start_date, next_day_first_hour_date_time, precipitation.volume / precipitation_hours * hours_until_next_day))
                temp_precipitations.append(Precipitation(precipitation.type_, next_day_first_hour_date_time, next_day_first_hour_date_time + datetime.timedelta(hours=next_day_precipitation_hours), precipitation.volume / precipitation_hours * next_day_precipitation_hours))
            else:
                temp_precipitations.append(precipitation)
        precipitations = temp_precipitations
        if rain := [precipitation for precipitation in precipitations if precipitation.type_ is PrecipitationType.RAIN]:
            distribute_precipitation_volume_(PrecipitationType.RAIN, rain)
        if snow := [precipitation for precipitation in precipitations if precipitation.type_ is PrecipitationType.SNOW]:
            distribute_precipitation_volume_(PrecipitationType.SNOW, snow)

    def fill_24h_instant_weathers(self):
        if not self.date:
            return

        for hour in range(24):
            if not self.get_instant_weathers_by_hour(hour):
                new_instant_weather = InstantWeather(datetime.datetime(self.date.year, self.date.month, self.date.day, hour, tzinfo=self.timezone))
                self.instant_weathers.insert(hour, new_instant_weather)

    def get_instant_weathers_by_hour(self, hour: int) -> list[InstantWeather]:
        return [instant_weather for instant_weather in self.instant_weathers if instant_weather.date_time and instant_weather.date_time.hour == hour]

    @classmethod
    def mean(
        cls,
        objects: Sequence,
        ratios: list[float] = None,
        attribute_names: Iterable[str] = ('sunrise', 'sunset', 'min_temperature', 'max_temperature')
    ) -> DayWeather:
        if not objects:
            return DayWeather()

        final_instant_weathers = []
        for hour in range(24):
            instant_weathers = [instant_weather for day_weather in objects if (instant_weather := next(iter(day_weather.get_instant_weathers_by_hour(hour)), None))]
            if instant_weathers:
                final_instant_weathers.append(InstantWeather.mean(instant_weathers, ratios))

        # noinspection PyTypeChecker
        day_weather: DayWeather = super().mean(objects, ratios, attribute_names)
        for object_ in objects:
            day_weather.date = object_.date
            day_weather.timezone = object_.timezone
            day_weather.instant_weathers = final_instant_weathers

        return day_weather

    def merge(self, other: DayWeather, left_priority=True) -> DayWeather:
        vars_copy = vars(self).copy()
        vars_copy.popitem()

        for k, v in vars_copy.items():
            if v in (None, []) or not left_priority:
                setattr(self, k, getattr(other, k, None))

        new_instant_weathers = []
        for hour in range(24):
            self_instant_weather = next(iter(self.get_instant_weathers_by_hour(hour)), None)
            other_instant_weather = next(iter(other.get_instant_weathers_by_hour(hour)), None)
            if self_instant_weather:
                if other_instant_weather:
                    new_instant_weathers.append(self_instant_weather.merge(other_instant_weather, left_priority))
                else:
                    new_instant_weathers.append(self_instant_weather)
            elif other_instant_weather:
                new_instant_weathers.append(other_instant_weather)
        self.instant_weathers = new_instant_weathers

        return self



import os
import re
from typing import Iterable

import aiohttp
import flanautils
from flanautils import Media, MediaType, OrderedSet, Source, return_if_first_empty

from flanaapis.exceptions import ResponseError, TwitterMediaNotFoundError

TWITTER_ENDPOINT_V1 = 'https://api.twitter.com/1.1/statuses/lookup.json'
TWITTER_ENDPOINT_V2 = 'https://api.twitter.com/2/tweets'
BEARER_TOKEN = os.environ['TWITTER_BEARER_TOKEN']
HEADERS = {'Authorization': f'Bearer {BEARER_TOKEN}'}


@return_if_first_empty(OrderedSet)
def find_tweet_ids(text: str | list[str]) -> OrderedSet[str]:
    if isinstance(text, list):
        text = ''.join(text)
    return OrderedSet(re.findall(r'atus/(\d+)', text))


def get_all_medias_from_tweet(tweet_medias_data: dict) -> OrderedSet[Media]:
    tweet_medias = OrderedSet()
    for tweet_media_data in tweet_medias_data:
        if tweet_media_data['type'] == 'photo':
            tweet_medias.add(Media(tweet_media_data['media_url_https'], MediaType.IMAGE, Source.TWITTER))
            continue

        try:
            variants = tweet_media_data['video_info']['variants']
        except KeyError:
            continue

        max_bitrate = float('-inf')
        max_bitrate_url = ''
        for variant in variants:
            try:
                bitrate = variant['bitrate']
            except KeyError:
                continue

            if bitrate > max_bitrate:
                max_bitrate = bitrate
                max_bitrate_url = variant['url']

        tweet_medias.add(Media(max_bitrate_url, MediaType.VIDEO if tweet_media_data['type'] == 'video' else MediaType.GIF, Source.TWITTER))

    return tweet_medias


async def get_medias(text: str) -> OrderedSet[Media]:
    medias: OrderedSet[Media] = OrderedSet()

    tweet_ids = find_tweet_ids(text)
    while referenced_tweet_ids := find_tweet_ids(await get_referenced_tweets(tweet_ids)) - tweet_ids:
        tweet_ids |= referenced_tweet_ids
    if not tweet_ids:
        return medias

    try:
        tweets_data = await get_tweets_data(TWITTER_ENDPOINT_V1, params={'id': ','.join(tweet_ids), 'tweet_mode': 'extended'})
    except aiohttp.ClientError:
        tweets_data = []

    for tweet_data in tweets_data:
        try:
            medias_data = tweet_data['extended_entities']['media']
        except KeyError:
            pass
        else:
            medias |= get_all_medias_from_tweet(medias_data)

    if not medias:
        raise TwitterMediaNotFoundError()

    return medias


@return_if_first_empty([])
async def get_referenced_tweets(tweet_ids: Iterable[str]) -> list[str]:
    try:
        tweets_data = await get_tweets_data(TWITTER_ENDPOINT_V2, params={'ids': ','.join(tweet_ids), 'expansions': 'attachments.media_keys', 'media.fields': 'media_key', 'tweet.fields': 'entities'})
    except ResponseError:
        return []

    referenced_tweets = []
    for tweet in tweets_data:
        try:
            urls = tweet['entities']['urls']
        except (KeyError, TypeError):
            continue

        for url in urls:
            try:
                referenced_tweets.append(url['expanded_url'])
            except KeyError:
                continue

    return referenced_tweets


async def get_tweets_data(url: str, params: dict) -> dict:
    data = await flanautils.get_request(url, params, headers=HEADERS)
    try:
        return data['data']
    except (TypeError, KeyError):
        return data



__all__ = [
    'add_daily_attributes_from_current_data_format',
    'create_instant_weather_by_data',
    'get_day_weathers_by_place',
    'get_hourly_precipitation',
    'get_hourly_precipitations',
    'get_weather_api_data'
]

import asyncio
import datetime
import os
from typing import Iterable, overload

import aiohttp
import flanautils

import flanaapis.geolocation.functions
import flanaapis.weather.functions
from flanaapis.exceptions import ResponseError
from flanaapis.geolocation.models import Place
from flanaapis.weather.models import DayPhases, DayWeather, InstantWeather, Precipitation, PrecipitationType

UNITS = 'metric'
LANGUAGE = 'es'
API_KEY = os.environ['OPEN_WEATHER_MAP_API_KEY']
BASE_ENDPOINT = 'https://api.openweathermap.org/data/2.5'
PAST_ENDPOINT = f'{BASE_ENDPOINT}/onecall/timemachine'
NEAR_FUTURE_ENDPOINT = f'{BASE_ENDPOINT}/forecast'
PRESENT_FUTURE_ENDPOINT = f'{BASE_ENDPOINT}/onecall'
MAX_RETRIES_PAST_REQUEST = 5


def add_daily_attributes_from_current_data_format(day_weathers: Iterable[DayWeather], data: dict, timezone: datetime.timezone):
    date_time = datetime.datetime.fromtimestamp(data['current']['dt'], tz=timezone)
    day_weather = flanautils.find(day_weathers, condition=lambda day_weather_: day_weather_.date == date_time.date())
    try:
        day_weather.sunrise = datetime.datetime.fromtimestamp(data['current']['sunrise'], timezone)
        day_weather.sunset = datetime.datetime.fromtimestamp(data['current']['sunset'], timezone)
    except KeyError:
        pass


def create_instant_weather_by_data(data: dict, timezone: datetime.timezone) -> InstantWeather:
    def format_old_open_weather_map_api_data() -> dict:
        for k, v in data['main'].items():
            data[k] = v
        data['clouds'] = sum(data['clouds'].values())
        data['wind_deg'] = data['wind']['deg']
        data['wind_gust'] = data['wind']['gust']
        data['wind_speed'] = data['wind']['speed']

        return data

    if 'main' in data:
        data = format_old_open_weather_map_api_data()

    return InstantWeather(
        date_time=datetime.datetime.fromtimestamp(data['dt'], tz=timezone),
        clouds=data.get('clouds'),
        description=data.get('weather', ({},))[0].get('description'),
        dew_point=data.get('dew_point'),
        humidity=data.get('humidity'),
        icon=data.get('weather', ({},))[0].get('icon'),
        precipitation_probability=pop * 100 if (pop := data.get('pop')) else pop,  # p -> %
        pressure=data.get('pressure'),
        temperature=data.get('temp'),
        temperature_feel=data.get('feels_like'),
        uvi=data.get('uvi'),
        visibility=visibility / 1000 if (visibility := data.get('visibility')) else visibility,  # visibility m -> km
        wind_degrees=data.get('wind_deg'),
        wind_gust=data.get('wind_gust'),
        wind_speed=wind_speed * 3.6 if (wind_speed := data.get('wind_speed')) else wind_speed  # wind_speed m/s -> km/h
    )


@overload
async def get_day_weathers_by_place(place: Place) -> tuple[InstantWeather, list[DayWeather]]:
    pass


@overload
async def get_day_weathers_by_place(place_query: str) -> tuple[InstantWeather, list[DayWeather]]:
    pass


@overload
async def get_day_weathers_by_place(latitude: float, longitude: float) -> tuple[InstantWeather, list[DayWeather]]:
    pass


async def get_day_weathers_by_place(latitude: float | str, longitude: float = None) -> tuple[InstantWeather, list[DayWeather]]:
    latitude, longitude = await flanaapis.geolocation.functions.ensure_coordinates(latitude, longitude)

    day_weathers = []

    past_days_data, present_future_data, near_future_data, timezone = await get_weather_api_data(latitude, longitude)
    current_weather = create_instant_weather_by_data(present_future_data['current'], timezone)
    precipitations: list[Precipitation] = []

    # ----- hourly data -----
    hourly_data = [hour_data for past_day_data in past_days_data for hour_data in past_day_data['hourly']][:-1]
    hourly_data += present_future_data['hourly']
    hourly_data += near_future_data['list']

    for hour_data in hourly_data:
        hour_dt = datetime.datetime.fromtimestamp(hour_data['dt'], tz=timezone)
        day_weather = flanautils.find(day_weathers, condition=lambda day_weather_: day_weather_.date == hour_dt.date())
        if not day_weather or hour_dt.date() != day_weather.date:
            day_weathers.append(day_weather := DayWeather(hour_dt.date(), timezone))

        if hourly_precipitations := get_hourly_precipitations(hour_dt, hour_data):
            precipitations.extend(hourly_precipitations)

        temp_day_weather = DayWeather()
        temp_day_weather.instant_weathers.append(create_instant_weather_by_data(hour_data, timezone))
        day_weather.merge(temp_day_weather)

    # ----- daily data -----
    for past_day_data in past_days_data:
        add_daily_attributes_from_current_data_format(day_weathers, past_day_data, timezone)

    add_daily_attributes_from_current_data_format(day_weathers, present_future_data, timezone)

    for future_day_data in present_future_data['daily']:
        day_dt = datetime.datetime.fromtimestamp(future_day_data['dt'], tz=timezone).replace(hour=0)
        day_weather = flanautils.find(day_weathers, condition=lambda day_weather_: day_weather_.date == day_dt.date())
        if not day_weather:
            day_weather = DayWeather(day_dt.date(), timezone)
            day_weathers.append(day_weather)

        day_weather.sunrise = datetime.datetime.fromtimestamp(future_day_data['sunrise'], timezone)
        day_weather.sunset = datetime.datetime.fromtimestamp(future_day_data['sunset'], timezone)
        day_weather.min_temperature = future_day_data['temp']['min']
        day_weather.max_temperature = future_day_data['temp']['max']
        if daily_rain := future_day_data.get('rain') or 0:
            precipitations.append(Precipitation(PrecipitationType.RAIN, day_dt, day_dt + datetime.timedelta(days=1), daily_rain))
        if daily_snow := future_day_data.get('snow') or 0:
            precipitations.append(Precipitation(PrecipitationType.SNOW, day_dt, day_dt + datetime.timedelta(days=1), daily_snow))

        temp_day_weather = DayWeather()
        # noinspection PyTypeChecker
        for phase_name, hour in DayPhases.items:
            phase_name = phase_name.lower()
            hour_dt = day_dt.replace(hour=hour)

            future_day_data_copy = future_day_data.copy()
            future_day_data_copy['dt'] = int(hour_dt.timestamp())
            future_day_data_copy['temp'] = future_day_data['temp'][phase_name]
            future_day_data_copy['feels_like'] = future_day_data['feels_like'][phase_name]
            if phase_name != 'day':
                future_day_data_copy['uvi'] = 0
            temp_day_weather.instant_weathers.append(create_instant_weather_by_data(future_day_data_copy, timezone))
            day_weather.merge(temp_day_weather)

    for day_weather in day_weathers:
        day_weather.distribute_precipitation_volume(precipitations)

    flanaapis.weather.functions.clear_past_precipitation_probability(day_weathers, timezone)

    return current_weather, day_weathers


def get_hourly_precipitation(date: datetime.datetime, data: dict, precipitation_type: PrecipitationType, last_hours: int) -> Precipitation:
    try:
        return Precipitation(precipitation_type, date - datetime.timedelta(hours=last_hours), date, data[precipitation_type.name.lower()][f'{last_hours}h'])
    except KeyError:
        pass


def get_hourly_precipitations(date: datetime.datetime, data: dict) -> list[Precipitation]:
    rain_last_hour = get_hourly_precipitation(date, data, PrecipitationType.RAIN, last_hours=1)
    rain_last_3_hours = get_hourly_precipitation(date, data, PrecipitationType.RAIN, last_hours=3)
    snow_last_hour = get_hourly_precipitation(date, data, PrecipitationType.SNOW, last_hours=1)
    snow_last_3_hours = get_hourly_precipitation(date, data, PrecipitationType.SNOW, last_hours=3)

    return [precipitation for precipitation in (rain_last_hour, rain_last_3_hours, snow_last_hour, snow_last_3_hours) if precipitation]


async def get_weather_api_data(latitude: float, longitude: float) -> tuple[list[dict], dict, dict, datetime.timezone]:
    parameters = {
        'lat': latitude,
        'lon': longitude,
        'units': UNITS,
        'lang': LANGUAGE,
        'appid': API_KEY
    }

    async with aiohttp.ClientSession() as session:
        past_days_data = []
        for days_to_the_past in reversed(range(6)):
            dt_query_param = datetime.datetime.now() - datetime.timedelta(days=days_to_the_past)
            for _ in range(MAX_RETRIES_PAST_REQUEST):
                try:
                    past_days_data.append(await flanautils.get_request(PAST_ENDPOINT, parameters | {'dt': int(dt_query_param.timestamp())}, session=session))
                    break
                except ResponseError:
                    await asyncio.sleep(1)

        present_future_data: dict = await flanautils.get_request(PRESENT_FUTURE_ENDPOINT, parameters, session=session)
        near_future_data: dict = await flanautils.get_request(NEAR_FUTURE_ENDPOINT, parameters, session=session)
        timezone = datetime.timezone(datetime.timedelta(seconds=present_future_data['timezone_offset']))

    return past_days_data, present_future_data, near_future_data, timezone



__all__ = ['create_instant_weather_by_data', 'get_weather_api_data', 'get_day_weathers_by_place']

import datetime
import os
from typing import overload

import flanautils

import flanaapis.geolocation.functions
import flanaapis.weather.functions
from flanaapis.exceptions import ResponseError
from flanaapis.geolocation.models import Place
from flanaapis.weather.models import DayWeather, InstantWeather, Precipitation, PrecipitationType

BASE_ENDPOINT = f'https://weather.visualcrossing.com/VisualCrossingWebServices/rest/services/timeline'


def create_instant_weather_by_data(data: dict, timezone: datetime.timezone) -> InstantWeather:
    return InstantWeather(
        date_time=datetime.datetime.fromtimestamp(data['datetimeEpoch'], timezone),
        clouds=data.get('cloudcover'),
        description=data.get('conditions'),
        dew_point=data.get('dew'),
        humidity=data.get('humidity'),
        icon=data.get('icon'),
        precipitation_probability=data.get('precipprob') or 0,  # %
        pressure=data.get('pressure'),
        temperature=data.get('temp'),
        temperature_feel=data.get('feelslike'),
        uvi=data.get('uvindex'),
        visibility=data.get('visibility'),  # km
        wind_degrees=data.get('winddir'),
        wind_gust=data.get('windgust'),
        wind_speed=data.get('windspeed')  # km/h
    )


@overload
async def get_day_weathers_by_place(place: Place) -> tuple[InstantWeather | None, list[DayWeather] | None]:
    pass


@overload
async def get_day_weathers_by_place(place_query: str) -> tuple[InstantWeather | None, list[DayWeather] | None]:
    pass


@overload
async def get_day_weathers_by_place(latitude: float, longitude: float) -> tuple[InstantWeather | None, list[DayWeather] | None]:
    pass


async def get_day_weathers_by_place(latitude: float | str, longitude: float = None) -> tuple[InstantWeather | None, list[DayWeather] | None]:
    latitude, longitude = await flanaapis.geolocation.functions.ensure_coordinates(latitude, longitude)

    try:
        api_data, timezone = await get_weather_api_data(latitude, longitude)
    except ResponseError:
        return None, None

    current_weather = create_instant_weather_by_data(api_data['currentConditions'], timezone)

    day_weathers = []
    for day_data in api_data['days']:
        precipitations: list[Precipitation] = []
        day_date = datetime.datetime.fromtimestamp(day_data['datetimeEpoch'], timezone).replace(hour=0)
        day_weather = DayWeather(
            date=day_date.date(),
            timezone=timezone,
            sunrise=datetime.datetime.fromtimestamp(day_data['sunriseEpoch'], timezone),
            sunset=datetime.datetime.fromtimestamp(day_data['sunsetEpoch'], timezone),
            min_temperature=day_data['tempmin'],
            max_temperature=day_data['tempmax']
        )

        # Daily precipitation volumes
        if day_rain_volume := day_data.get('precip') or 0:
            precipitations.append(Precipitation(PrecipitationType.RAIN, day_date, day_date + datetime.timedelta(days=1), day_rain_volume))
        if day_snow_volume := day_data.get('snow') or 0:
            precipitations.append(Precipitation(PrecipitationType.SNOW, day_date, day_date + datetime.timedelta(days=1), day_snow_volume * 10))  # snow cm -> mm

        # Hourly precipitation volumes
        for hour_data in day_data.get('hours', ()):
            day_weather.instant_weathers.append(create_instant_weather_by_data(hour_data, timezone))
            hour_date = datetime.datetime.fromtimestamp(hour_data['datetimeEpoch'], timezone)
            if hour_rain_volume := hour_data.get('precip') or 0:
                precipitations.append(Precipitation(PrecipitationType.RAIN, hour_date, hour_date + datetime.timedelta(hours=1), hour_rain_volume))
            if hour_snow_volume := hour_data.get('snow') or 0:
                precipitations.append(Precipitation(PrecipitationType.SNOW, hour_date, hour_date + datetime.timedelta(hours=1), hour_snow_volume * 10))  # snow cm -> mm

        day_weather.distribute_precipitation_volume(precipitations)

        day_weathers.append(day_weather)

    flanaapis.weather.functions.clear_past_precipitation_probability(day_weathers, timezone)

    return current_weather, day_weathers


async def get_weather_api_data(latitude: float, longitude: float) -> tuple[dict, datetime.timezone]:
    now = datetime.datetime.now()
    start_date = now - datetime.timedelta(days=5)
    end_date = now + datetime.timedelta(days=15)
    parameters = {
        'key': os.environ['VISUAL_CROSSING_API_KEY'],
        'unitGroup': 'metric',
        'lang': 'es'
    }

    api_data = await flanautils.get_request(f'{BASE_ENDPOINT}/{latitude},{longitude}/{int(start_date.timestamp())}/{int(end_date.timestamp())}', parameters)
    timezone = datetime.timezone(datetime.timedelta(hours=api_data['tzoffset']))

    return api_data, timezone



from __future__ import annotations  # todo0 remove in 3.11

import base64
from enum import Enum

from fastapi import APIRouter
from flanautils import Media
from pydantic import BaseModel

from flanaapis.scraping import instagram, tiktok, twitter, youtube

router = APIRouter(prefix='/medias')


def media_to_dict(media: Media) -> dict:
    response_media_vars = {}

    media_vars = media.to_dict()
    for k, v in media_vars.items():
        k = k.rstrip('_')

        match v:
            case bytes():
                v = base64.b64encode(v)
            case Enum():
                v = v.name.lower()
            case Media():
                v = media_to_dict(v)

        response_media_vars[k] = v

    return response_media_vars


class Input(BaseModel):
    text: str


class MediaOutput(BaseModel):
    url: str = None
    bytes: bytes = None
    type: str = None
    source: str = None
    title: str = None
    author: str = None
    album: str = None
    song_info: MediaOutput = None


@router.post('/', response_model=list[MediaOutput], response_model_exclude_defaults=True, response_model_exclude_unset=True)
async def get_medias(input_: Input):
    return [media_to_dict(media) for media in (
        *await instagram.get_medias(input_.text),
        *await tiktok.get_medias(input_.text),
        *await twitter.get_medias(input_.text),
        *await youtube.get_medias(input_.text)
    )]



import asyncio
import concurrent.futures
import re

import flanautils
from TikTokApi import TikTokApi
from flanautils import Media, MediaType, OrderedSet, Source

from flanaapis.exceptions import TikTokMediaNotFoundError
from flanaapis.scraping import constans


def find_download_urls(text: str) -> list[str]:
    partial_download_urls = re.findall(r'web\.tiktok\.com.+=&vr=', text)
    return [f'https://v16-{partial_download_url}' for partial_download_url in partial_download_urls]


def find_tiktok_ids(text: str) -> OrderedSet[str]:
    return OrderedSet(re.findall(r'v(?:ideo)?/(\d+)', text))


async def find_tiktok_ids_by_mobile_urls(text: str) -> OrderedSet[str]:
    mobile_ids = re.findall(r'vm\.tiktok\.com/(\w+)', text)
    mobile_tiktok_urls = [f'https://vm.tiktok.com/{mobile_id}/' for mobile_id in mobile_ids]
    tiktok_urls_gen = [str((await flanautils.get_request(mobile_tiktok_url, headers={'User-Agent': constans.USER_AGENT}, return_response=True)).url) for mobile_tiktok_url in mobile_tiktok_urls]
    return find_tiktok_ids(''.join(tiktok_urls_gen))


def get_media_dict_by_download_url(download_url: str, song_data: dict = None) -> dict:
    song_info = Media(
        song_data['playUrl'],
        MediaType.AUDIO,
        Source.TIKTOK,
        song_data['title'],
        song_data['authorName'],
        song_data['album']
    ) if song_data else None

    return Media(download_url, MediaType.VIDEO, Source.TIKTOK, song_info=song_info).to_dict()


def get_media_dict_by_id(tiktok_id: str) -> dict:
    api: TikTokApi = TikTokApi.get_instance()

    tiktok_data = api.get_tiktok_by_id(tiktok_id)
    video_data = tiktok_data['itemInfo']['itemStruct']['video']
    song_data = tiktok_data['itemInfo']['itemStruct']['music']
    try:
        download_url = video_data['playAddr']
    except KeyError:
        download_url = video_data['downloadAddr']

    return get_media_dict_by_download_url(download_url, song_data)


async def get_medias(text: str) -> OrderedSet[Media]:
    medias: OrderedSet[Media] = OrderedSet()

    tiktok_ids = find_tiktok_ids(text)
    tiktok_ids += await find_tiktok_ids_by_mobile_urls(text)
    download_urls = find_download_urls(text)
    if not tiktok_ids and not download_urls:
        return medias

    with concurrent.futures.ProcessPoolExecutor(max_workers=3) as pool:
        for tiktok_id in tiktok_ids:
            result_dict = await asyncio.get_running_loop().run_in_executor(pool, get_media_dict_by_id, tiktok_id)
            medias.add(Media.from_dict(result_dict, lazy=False))

        for download_url in download_urls:
            result_dict = await asyncio.get_running_loop().run_in_executor(pool, get_media_dict_by_download_url, download_url)
            medias.add(Media.from_dict(result_dict, lazy=False))

    if not medias:
        raise TikTokMediaNotFoundError

    return medias



from flanautils.exceptions import NotFoundError, ResponseError


class InstagramLoginError(ResponseError):
    pass


class MediaNotFoundError(NotFoundError):
    source: str


class PlaceNotFoundError(NotFoundError):
    pass


class InstagramMediaNotFoundError(MediaNotFoundError):
    source = 'Instagram'


class TikTokMediaNotFoundError(MediaNotFoundError):
    source = 'TikTok'


class TwitterMediaNotFoundError(MediaNotFoundError):
    source = 'Twitter'



from __future__ import annotations  # todo0 remove in 3.11

from flanautils import FlanaBase, OrderedSet


class Place(FlanaBase):
    """Simple class for manage Place data."""

    def __init__(
        self,
        name: str = None,
        latitude: float = None,
        longitude: float = None,
        country: str = None,
        country_code: str = None,
        state: str = None,
        state_district: str = None,
        province: str = None,
        county: str = None,
        city: str = None,
        town: str = None,
        borough: str = None,
        postcode: str = None,
        neighbourhood: str = None,
        road: str = None,
        number: str = None,
        amenity: str = None
    ):
        self._name = ', '.join(OrderedSet(name.split(', '))) if name else name
        self.latitude = float(latitude) if latitude else latitude
        self.longitude = float(longitude) if longitude else longitude
        self.country = country
        self.country_code = country_code
        self.state = state
        self.state_district = state_district
        self.province = province
        self.county = county
        self.city = city
        self.town = town
        self.borough = borough
        self.postcode = postcode
        self.neighbourhood = neighbourhood
        self.road = road
        self.number = number
        self.amenity = amenity

    def __str__(self):
        return self.name

    @property
    def name(self):
        address_parts = [address_part.strip() if len(address_part) <= 10 else f'{address_part[:10].strip()}...' for address_part in (self.town, self.city, self.county, self.state, self.country) if address_part]
        return ', '.join(address_parts) if address_parts else self._name

    def distance_to(self, place: Place) -> float:
        """Calculate the distance between two places."""

        return ((place.latitude - self.latitude) ** 2 + (place.longitude - self.longitude) ** 2) ** (1 / 2)



import os

import flanautils

os.environ |= flanautils.find_environment_variables('../.env')

import sys

import uvicorn
from fastapi import FastAPI

import flanaapis.geolocation.routes
import flanaapis.scraping.routes
import flanaapis.weather.routes

sub_app = FastAPI()
sub_app.include_router(flanaapis.geolocation.routes.router)
sub_app.include_router(flanaapis.scraping.routes.router)
sub_app.include_router(flanaapis.weather.routes.router)

app = FastAPI()
app.mount('/flanaapis', sub_app)

if __name__ == '__main__':
    try:
        host = sys.argv[1]
    except IndexError:
        host = os.environ.get('FLANAAPIS_HOST')
    try:
        port = sys.argv[2]
    except IndexError:
        port = os.environ.get('FLANAAPIS_PORT')

    uvicorn.run('main:app', host=host, port=int(port))



import html
import os
import re
from typing import Iterable

import playwright
import playwright.async_api
from flanautils import Media, MediaType, OrderedSet, Source

from flanaapis.exceptions import InstagramMediaNotFoundError
from flanaapis.scraping import constans

INSTAGRAM_BASE_URL = 'https://www.instagram.com/'
INSTAGRAM_LOGIN_URL = INSTAGRAM_BASE_URL + 'accounts/login/ajax/'
INSTAGRAM_CONTENT_PATH = 'p/'

cookies = None


def find_instagram_ids(text: str) -> OrderedSet[str]:
    return OrderedSet(re.findall(r'/(?:p|reel|tv)/(.{11})', text))


def find_media_mark(text: str) -> str:
    try:
        return re.findall(r'[\w.]+\.(?:jpg|webp|mp4)\?', text)[0]
    except IndexError:
        return ''


def find_media_urls(text: str) -> list[str]:
    return re.findall(r'https(?:(?!https).)*?sid=\w{6}', text)


async def get_medias(text: str) -> OrderedSet[Media]:
    global cookies

    medias: OrderedSet[Media] = OrderedSet()

    if not (instagram_urls := make_instagram_urls(find_instagram_ids(text))):
        return medias

    async with playwright.async_api.async_playwright() as p:
        async with await p.chromium.launch() as browser:
            context = await browser.new_context(user_agent=constans.USER_AGENT_2, locale='es-ES')
            page: playwright.async_api.Page = await context.new_page()

            if cookies:
                # noinspection PyTypeChecker
                await context.add_cookies(cookies)
            else:
                await login(page)
                cookies = await context.cookies()

            for instagram_url in instagram_urls:
                await page.goto(instagram_url)
                await page.wait_for_load_state('networkidle')
                button = page.locator('button[aria-label=Siguiente]')
                while await button.count():
                    await button.click()
                await page.wait_for_load_state('networkidle')
                html_content = html.unescape(await page.content())

                new_medias = select_content_urls(find_media_urls(html_content))
                for media in new_medias:
                    response = await context.request.fetch(media.url)
                    media.bytes_ = await response.body()
                medias |= new_medias

    if not medias:
        raise InstagramMediaNotFoundError

    return medias


async def login(page: playwright.async_api.Page):
    await page.goto(INSTAGRAM_BASE_URL)
    button = page.locator("'Permitir cookies necesarias y opcionales'")
    if await button.count():
        await button.click()
    await page.fill('input[name=username]', os.environ['INSTAGRAM_USERNAME'])
    await page.fill('input[name=password]', os.environ['INSTAGRAM_PASSWORD'])
    await page.click("'Entrar'")
    await page.wait_for_load_state('networkidle')
    button = page.locator("'Guardar información'")
    if await button.count():
        await button.click()
    await page.wait_for_load_state('networkidle')
    button = page.locator("'Ahora no'")
    if await button.count():
        await button.click()


def make_instagram_urls(ids: Iterable[str]) -> list[str]:
    return [f'{INSTAGRAM_BASE_URL}{INSTAGRAM_CONTENT_PATH}{id}' for id in ids]


def select_content_urls(media_urls: list[str]) -> OrderedSet[Media]:
    last_url = ''
    final_urls = OrderedSet()
    thumbnail_urls = OrderedSet()
    for media_url in media_urls:
        if not re.findall(r'e0&cb.*cache', media_url) and '.mp4?efg' not in media_url:
            last_url = ''
            continue

        if '.mp4?' in media_url and 'jpg' in last_url:
            thumbnail_urls.add(last_url)
        final_urls.add(media_url)

        last_url = media_url

    final_urls -= thumbnail_urls

    return OrderedSet(Media(final_url, MediaType.VIDEO if '.mp4?' in final_url else MediaType.IMAGE, Source.INSTAGRAM) for final_url in final_urls)



from flanautils import FlanaEnum


class WeatherEmoji(FlanaEnum):
    ZOOM_IN = '🔎➕'
    ZOOM_OUT = '🔎➖'
    LEFT = '⬅'
    RIGHT = '➡'
    TEMPERATURE = '🌡'
    TEMPERATURE_FEEL = '🖐🌡'
    CLOUDS = '☁☁'
    VISIBILITY = '〰👀'
    UVI = '↯☀'
    HUMIDITY = '💦'
    PRECIPITATION_PROBABILITY = '🎲☔'
    PRECIPITATION_VOLUME = '🪣🌧'
    PRESSURE = '🌪⇊'
    WIND_SPEED = '🌬⇉'



import re
import urllib.parse
from typing import AsyncIterable

import playwright
from flanaapis.geolocation.models import Place
from playwright.async_api import async_playwright


async def find_place(place_query: str) -> Place | None:
    async with async_playwright() as p:
        async with await p.chromium.launch() as browser:
            try:
                page: playwright.async_api.Page = await browser.new_page()
                await page.goto(f"https://www.google.es/maps/search/{'+'.join(place_query.split())}")
                button = page.locator("'Acepto'")
                if await button.count():
                    await button.click()
                while '@' not in page.url:
                    await page.wait_for_event('framenavigated')

                await page.wait_for_load_state('domcontentloaded')
                if await page.query_selector("'Google Maps no encuentra'"):
                    return
            except playwright.async_api.TimeoutError:
                return

            try:
                place_name, latitude, longitude = re.findall(r'(?:place|search)/\d*(.+)/@([\d.-]+),([\d.-]+)', page.url)[0]
            except IndexError:
                return

            place_name = urllib.parse.unquote(place_name.replace('+', ' ').strip())

            return Place(place_name, latitude, longitude)


async def find_place_showing_progress(place_query: str) -> AsyncIterable[str | Place | None]:
    yield 'Abriendo navegador...'
    async with async_playwright() as p:
        async with await p.chromium.launch() as browser:
            try:
                page: playwright.async_api.Page = await browser.new_page()
                yield 'Dirigiéndome a google.es/maps...'
                await page.goto(f"https://www.google.es/maps/search/{'+'.join(place_query.split())}")

                if await (button := page.locator("button:has-text('Acepto')")).count() or await (button := page.locator("button:has-text('Aceptar todo')")).count():
                    yield 'Aceptando consentimiento de privacidad...'
                    await button.first.click()

                yield 'Rebuscando coordenadas en la página...'
                while '@' not in page.url:
                    await page.wait_for_event('framenavigated')

                await page.wait_for_load_state('domcontentloaded')
                if await page.query_selector("'Google Maps no encuentra'"):
                    yield
                    return
            except playwright.async_api.TimeoutError:
                yield
                return

            try:
                place_name, latitude, longitude = re.findall(r'(?:place|search)/\d*(.+)/@([\d.-]+),([\d.-]+)', page.url)[0]
            except IndexError:
                yield
                return

            place_name = urllib.parse.unquote(place_name.replace('+', ' ').strip())

            yield Place(place_name, latitude, longitude)



from typing import overload

import flanautils
from flanaapis.geolocation.models import Place


@overload
async def find_place(place_query: str, near_to_place: Place = None) -> Place | None:
    pass


@overload
async def find_place(place_query: str, near_to_latitude: float = None, near_to_longitude: float = None) -> Place | None:
    pass


async def find_place(place_query: str, near_to_latitude: float = None, near_to_longitude: float = None) -> Place | None:
    match near_to_latitude, near_to_longitude:
        case Place() as center_place, _:
            pass
        case int() | float(), int() | float():
            center_place = Place('', near_to_latitude, near_to_longitude)
        case _:
            center_place = None

    if center_place:
        nearest_place = None
        shortest_distance = float('inf')
        for place in await find_places(place_query):
            if (distance := place.distance_to(center_place)) < shortest_distance:
                shortest_distance = distance
                nearest_place = place

        if nearest_place:
            return nearest_place
    else:
        return next(iter(await find_places(place_query)), None)


async def find_places(place_query: str) -> list[Place]:
    places_data: list[dict] = await flanautils.get_request(f'https://nominatim.openstreetmap.org/search',
                                                           {'q': place_query,
                                                            'format': 'jsonv2',
                                                            'accept-language': 'es-ES,es,en',
                                                            'addressdetails': True})

    places = []
    for place_data in places_data:
        place = Place(place_data.get('display_name'), place_data.get('lat'), place_data.get('lon'))

        if 'address' in place_data:
            place.country = place_data['address'].get('country')
            place.country_code = place_data['address'].get('country_code')
            place.state = place_data['address'].get('state')
            place.state_district = place_data['address'].get('state_district')
            place.province = place_data['address'].get('province')
            place.county = place_data['address'].get('county')
            place.city = place_data['address'].get('city')
            place.town = place_data['address'].get('town')
            place.borough = place_data['address'].get('borough')
            place.postcode = place_data['address'].get('postcode')
            place.neighbourhood = place_data['address'].get('neighbourhood')
            place.road = place_data['address'].get('road')
            place.number = place_data['address'].get('house_number')
            place.amenity = place_data['address'].get('amenity')

        places.append(place)

    return places



from fastapi import APIRouter
from pydantic import BaseModel

from flanaapis.geolocation import functions, open_street_map

router = APIRouter()


class PlaceOutput(BaseModel):
    name: str = None
    latitude: float = None
    longitude: float = None
    country: str = None
    country_code: str = None
    state: str = None
    state_district: str = None
    county: str = None
    city: str = None
    borough: str = None
    postcode: str = None
    neighbourhood: str = None
    road: str = None
    number: str = None
    amenity: str = None


@router.get("/place", response_model=PlaceOutput, response_model_exclude_defaults=True, response_model_exclude_unset=True)
async def find_place(query: str, near_to_latitude: float = None, near_to_longitude: float = None, fast: bool = False):
    match near_to_latitude, near_to_longitude:
        case [float(), _] | [_, float()]:
            return (await open_street_map.find_place(query, near_to_latitude, near_to_longitude)).to_dict()
        case _ if fast:
            return (await open_street_map.find_place(query, near_to_latitude, near_to_longitude)).to_dict()
        case _:
            return (await functions.find_place(query)).to_dict()


# noinspection PyUnusedLocal
@router.get("/places", response_model=list[PlaceOutput], response_model_exclude_defaults=True, response_model_exclude_unset=True)
async def find_places(query: str, fast: bool = False):
    return [place.to_dict() for place in await open_street_map.find_places(query)]


@router.get("/timezone")
async def find_places(query: str, fast: bool = False):
    return await functions.find_timezone(query, fast) or {}



from flanaapis.weather.constants import *
from flanaapis.weather.functions import *
from flanaapis.weather.google import *
from flanaapis.weather.models import *
from flanaapis.weather.open_weather_map import *
from flanaapis.weather.visual_crossing import *



USER_AGENT = 'Instagram 123.0.0.21.114 (iPhone; CPU iPhone OS 11_4 like Mac OS X; en_US; en-US; scale=2.00; 750x1334) AppleWebKit/605.1.15'
USER_AGENT_2 = 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/102.0.0.0 Safari/537.36'



import os
from typing import AsyncIterator, overload

import flanautils

from flanaapis.exceptions import PlaceNotFoundError
from flanaapis.geolocation import google_maps, open_street_map
from flanaapis.geolocation.models import Place

TIMEZONE_API_KEY = os.environ['TIMEZONEDB_API_KEY']
TIMEZONE_BASE_ENDPOINT = 'https://api.timezonedb.com/v2.1/get-time-zone'


@overload
async def ensure_coordinates(place_query: str, fast: bool = False) -> tuple[float, float]:
    pass


@overload
async def ensure_coordinates(latitude: float, longitude: float = None, fast: bool = False) -> tuple[float, float]:
    pass


async def ensure_coordinates(latitude: float | str, longitude: float = None, fast: bool = False) -> tuple[float, float]:
    match latitude, longitude:
        case Place() as place, _:
            latitude = place.latitude
            longitude = place.longitude
        case str(), _ if coordinates := flanautils.find_coordinates(latitude):
            latitude, longitude = coordinates
        case str(place_query), _ if place := await find_place(place_query, fast):
            latitude = place.latitude
            longitude = place.longitude
        case str(), _:
            raise PlaceNotFoundError

    return latitude, longitude


@overload
async def find_timezone(place_query: str, fast: bool = False) -> dict | None:
    pass


@overload
async def find_timezone(latitude: float, longitude: float = None, fast: bool = False) -> dict | None:
    pass


async def find_timezone(latitude: float | str, longitude: float = None, fast: bool = False) -> dict | None:
    latitude, longitude = await ensure_coordinates(latitude, longitude, fast)
    parameters = {
        'key': TIMEZONE_API_KEY,
        'format': 'json',
        'by': 'position',
        'lat': latitude,
        'lng': longitude
    }
    timezone_data = await flanautils.get_request(TIMEZONE_BASE_ENDPOINT, params=parameters)

    if timezone_data['status'] == 'OK':
        return timezone_data


async def find_place(place_query: str, fast: bool = False) -> Place | None:
    if not fast and (partial_place := await google_maps.find_place(place_query)):
        return await open_street_map.find_place(f'{partial_place.latitude}, {partial_place.longitude}')
    else:
        return await open_street_map.find_place(place_query)


# noinspection PyUnusedLocal
async def find_places(place_query: str, fast: bool) -> list[Place]:
    return await open_street_map.find_places(place_query)


async def find_place_showing_progress(place_query: str) -> AsyncIterator[str | Place | None]:
    result: str | Place | None = None
    async for result in google_maps.find_place_showing_progress(place_query):
        match result:
            case None:
                break
            case str(message):
                yield message

    if result:
        yield await open_street_map.find_place(f'{result.latitude}, {result.longitude}')
    else:
        yield 'Google maps no ha encontrado nada. Buscando en openstreetmap.org...'
        yield await open_street_map.find_place(place_query)
